@@ -1,14 +1,13 @@
 use anyhow::{Context, Result};
-use glob::Pattern;
+use linter::Linter;
 use log::debug;
-use render::render_lint_messages;
+use render::{print_error, render_lint_messages};
 use std::collections::HashMap;
 use std::io::Write;
 use std::path::PathBuf;
-use std::process::exit;
 use std::sync::{Arc, Mutex};
-use std::thread;
 use std::{collections::HashSet, process::Command};
+use std::{io, thread};
 use structopt::StructOpt;
 
 mod lint_config;
@@ -16,10 +15,29 @@ mod lint_message;
 mod linter;
 mod render;
 
-use lint_config::LintConfig;
 use lint_message::LintMessage;
 use render::PrintedLintErrors;
 
+use crate::lint_config::get_linters_from_config;
+
+fn get_paths_cmd_files(paths_cmd: String) -> Result<Vec<String>> {
+    debug!("Running paths_cmd: {}", paths_cmd);
+    let output = Command::new("sh")
+        .arg("-c")
+        .arg(paths_cmd)
+        .output()
+        .context("failed to run provided paths_cmd")?;
+
+    let files = std::str::from_utf8(&output.stdout).context("failed to parse paths_cmd output")?;
+    let files = files
+        .lines()
+        .map(|s| s.to_string())
+        .collect::<HashSet<String>>();
+    let mut files = files.into_iter().collect::<Vec<String>>();
+    files.sort();
+    Ok(files)
+}
+
 fn get_changed_files() -> Result<Vec<String>> {
     // Retrieve changed files in current commit
     let commit_files = Command::new("git")
@@ -49,9 +67,9 @@ fn get_changed_files() -> Result<Vec<String>> {
 
     let working_tree_files_str = std::str::from_utf8(&working_tree_files.stdout)?;
     let working_tree_files: HashSet<String> = working_tree_files_str
-        .split("\n")
-        .map(|x| x.to_string())
+        .lines()
         .filter(|line| !line.is_empty())
+        .map(|x| x.to_string())
         .collect();
     let mut all_changed_files: Vec<String> = commit_files
         .union(&working_tree_files)
@@ -63,14 +81,6 @@ fn get_changed_files() -> Result<Vec<String>> {
     Ok(all_changed_files)
 }
 
-fn write_matches_to_file(matched_files: Vec<String>) -> Result<tempfile::NamedTempFile> {
-    let file = tempfile::NamedTempFile::new()?;
-    for matched_file in matched_files {
-        writeln!(&file, "{}", matched_file)?;
-    }
-    Ok(file)
-}
-
 fn group_lints_by_file(
     all_lints: &mut HashMap<PathBuf, Vec<LintMessage>>,
     lints: Vec<LintMessage>,
@@ -83,81 +93,6 @@ fn group_lints_by_file(
     });
 }
 
-struct Linter {
-    name: String,
-    patterns: Vec<Pattern>,
-    commands: Vec<String>,
-}
-
-impl Linter {
-    fn get_matches(&self, files: &Vec<String>) -> Vec<String> {
-        files
-            .iter()
-            .filter(|name| self.patterns.iter().any(|pattern| pattern.matches(name)))
-            .map(|name| name.clone())
-            .collect()
-    }
-
-    fn run_command(&self, filenames_to_lint: tempfile::NamedTempFile) -> Result<Vec<LintMessage>> {
-        let file_path = filenames_to_lint
-            .path()
-            .to_str()
-            .ok_or(anyhow::Error::msg("tempfile corrupted"))?;
-
-        let (program, arguments) = self.commands.split_at(1);
-        let arguments: Vec<String> = arguments
-            .iter()
-            .map(|arg| arg.replace("{{PATHSFILE}}", file_path))
-            .collect();
-
-        debug!("Running: {} {}", program[0], arguments.join(" "));
-        let command = Command::new(&program[0]).args(arguments).output()?;
-
-        if !&command.status.success() {
-            let stderr = std::str::from_utf8(&command.stderr)?.to_owned();
-            return Err(anyhow::Error::msg(format!(
-                "lint adapter for '{}' failed with non-zero exit code",
-                self.name
-            )))
-            .with_context(|| stderr);
-        }
-        let stdout_str = std::str::from_utf8(&command.stdout)?;
-        let lints: Vec<LintMessage> = stdout_str
-            .split("\n")
-            .filter(|line| !line.is_empty())
-            .map(|line| serde_json::from_str(line).map_err(|a| anyhow::Error::msg(a.to_string())))
-            .collect::<Result<_>>()
-            .context(format!(
-                "Failed to deserialize output for lint adapter: '{}'",
-                self.name
-            ))?;
-
-        Ok(lints)
-    }
-
-    fn run(&self, files: &Vec<String>) -> Result<Vec<LintMessage>> {
-        let matches = self.get_matches(files);
-        debug!("Linter '{}' matched files: {:#?}", self.name, matches);
-        if matches.is_empty() {
-            return Ok(Vec::new());
-        }
-        let file = write_matches_to_file(matches)?;
-        self.run_command(file)
-    }
-}
-
-fn patterns_from_strs(pattern_strs: &Vec<String>) -> Result<Vec<Pattern>> {
-    pattern_strs
-        .iter()
-        .map(|pattern_str| {
-            Pattern::new(pattern_str).map_err(|err| {
-                anyhow::Error::msg(err)
-                    .context("Could not parse pattern from linter configuration.")
-            })
-        })
-        .collect::<Result<Vec<Pattern>>>()
-}
-
 fn apply_patches(lint_messages: &HashMap<PathBuf, Vec<LintMessage>>) -> Result<()> {
     for (path, lint_messages) in lint_messages {
         for lint_message in lint_messages {
@@ -173,43 +108,76 @@ fn apply_patches(lint_messages: &HashMap<PathBuf, Vec<LintMessage>>) -> Result<(
 }
 
 #[derive(Debug, StructOpt)]
-#[structopt(name = "example", about = "An example of StructOpt usage.")]
+#[structopt(name = "lintrunner", about = "A lint runner")]
 struct Opt {
-    #[structopt(long = "config", default_value = ".lintrunner.toml")]
-    config: String,
-
-    #[structopt(short = "v", long = "verbose")]
+    #[structopt(short, long)]
     verbose: bool,
 
-    #[structopt(short = "a", long = "apply-patches")]
+    /// Path to a toml file defining which linters to run
+    #[structopt(long, default_value = ".lintrunner.toml")]
+    config: String,
+
+    /// If set, any suggested patches will be applied
+    #[structopt(short, long)]
     apply_patches: bool,
+
+    /// Shell command that returns new-line separated paths to lint
+    /// (e.g. --paths-cmd 'git ls-files path/to/project')
+    #[structopt(long)]
+    paths_cmd: Option<String>,
+
+    /// Comma-separated list of linters to skip (e.g. --skip CLANGFORMAT,NOQA")
+    #[structopt(long)]
+    skip: Option<String>,
+
+    /// Comma-separated list of linters to run (opposite of --skip)
+    #[structopt(long)]
+    take: Option<String>,
+
+    #[structopt(subcommand)]
+    cmd: Option<SubCommand>,
 }
 
-fn main() -> Result<()> {
-    let opt = Opt::from_args();
-    let log_level = if opt.verbose {
-        log::LevelFilter::Debug
-    } else {
-        log::LevelFilter::Info
-    };
-    env_logger::Builder::new().filter_level(log_level).init();
+#[derive(StructOpt, Debug)]
+enum SubCommand {
+    /// Perform first-time setup for linters
+    Init {
+        /// If set, do not actually execute initialization commands, just print them
+        #[structopt(long, short)]
+        dry_run: bool,
+    },
+}
 
-    let config_path = PathBuf::from(opt.config);
-    let lint_config = LintConfig::new(&config_path)?;
+fn do_init(linters: Vec<Linter>, dry_run: bool) -> Result<i32> {
+    debug!(
+        "Initializing linters: {:?}",
+        linters.iter().map(|l| &l.name).collect::<Vec<_>>()
+    );
 
-    let mut linters = Vec::new();
-    for config in lint_config.linters {
-        let patterns = patterns_from_strs(&config.patterns)?;
-        linters.push(Linter {
-            name: config.name,
-            patterns,
-            commands: config.args,
-        });
+    for linter in linters {
+        linter.init(dry_run)?;
     }
 
+    Ok(0)
+}
+
+fn do_lint(
+    linters: Vec<Linter>,
+    paths_cmd: Option<String>,
+    should_apply_patches: bool,
+) -> Result<i32> {
+    debug!(
+        "Running linters: {:?}",
+        linters.iter().map(|l| &l.name).collect::<Vec<_>>()
+    );
+
     // Too lazy to learn rust's fancy concurrent programming stuff, just spawn a thread per linter and join them.
     let all_lints = Arc::new(Mutex::new(HashMap::new()));
-    let files = Arc::new(get_changed_files()?);
+    let files = match paths_cmd {
+        Some(paths_cmd) => get_paths_cmd_files(paths_cmd)?,
+        None => get_changed_files()?,
+    };
+    let files = Arc::new(files);
 
     debug!("Linting files: {:#?}", files);
 
@@ -234,14 +202,69 @@ fn main() -> Result<()> {
     let all_lints = all_lints.lock().unwrap();
     let did_print = render_lint_messages(&all_lints)?;
     match did_print {
-        PrintedLintErrors::No => {
-            exit(0);
-        }
+        PrintedLintErrors::No => Ok(0),
         PrintedLintErrors::Yes => {
-            if opt.apply_patches {
+            if should_apply_patches {
                 apply_patches(&all_lints)?;
             }
-            exit(1);
+            Ok(1)
         }
     }
 }
+
+fn do_main() -> Result<i32> {
+    let opt = Opt::from_args();
+    let log_level = if opt.verbose {
+        log::LevelFilter::Debug
+    } else {
+        log::LevelFilter::Info
+    };
+    env_logger::Builder::new().filter_level(log_level).init();
+
+    let config_path = PathBuf::from(opt.config);
+    let skipped_linters = opt.skip.map(|linters| {
+        linters
+            .split(',')
+            .map(|linter_name| linter_name.to_string())
+            .collect::<HashSet<_>>()
+    });
+    let taken_linters = opt.take.map(|linters| {
+        linters
+            .split(',')
+            .map(|linter_name| linter_name.to_string())
+            .collect::<HashSet<_>>()
+    });
+
+    let linters = get_linters_from_config(&config_path, skipped_linters, taken_linters)?;
+
+    match opt.cmd {
+        Some(SubCommand::Init { dry_run }) => {
+            // Just run initialization commands, don't actually lint.
+            do_init(linters, dry_run)
+        }
+        None => {
+            // Default command is to just lint.
+            do_lint(linters, opt.paths_cmd, opt.apply_patches)
+        }
+    }
+}
+
+fn main() {
+    let code = match do_main() {
+        Ok(code) => code,
+        Err(err) => {
+            print_error(&err)
+                .context("failed to print exit error")
+                .unwrap();
+            1
+        }
+    };
+
+    // Flush the output before exiting, in case there is anything left in the buffers.
+    drop(io::stdout().flush());
+    drop(io::stderr().flush());
+
+    // exit() abruptly ends the process while running no destructors. We should
+    // make sure that nothing is alive before running this.
+    std::process::exit(code);
+}
